



pycc -h
IF %ERRORLEVEL% NEQ 0 exit 1
numba -h
IF %ERRORLEVEL% NEQ 0 exit 1
numba -s
IF %ERRORLEVEL% NEQ 0 exit 1
python -m numba.tests.test_runtests
IF %ERRORLEVEL% NEQ 0 exit 1
python -m numba.runtests -m -b
IF %ERRORLEVEL% NEQ 0 exit 1
exit 0
